from .pyappscan import PyAppscan

def create_engine(url, username=None, password=None, key='AppscanEnterpriseUser', apikey=None, secretkey=None, verify=True):
	"""Convenience function for importing all modules.

	This function is in the base of the module and will return a properly initialized object namespace with which you can access all appscan methods. This abstracts the state management and login management of appscan. Usage:
	
	Example:
			>>> from pyappscan import create_engine
			# imported
			>>> appscan = create_engine(
			...		url="https://appscan.domain.com/ase/api/", 
			...		apikey="AEF8137319C8", 
			...		secretkey="XXXXXXXX"
			... )
			
	Args:
		url (str): Base URL for appscan API
		username (str, optional): Username to log in with; used in conjunction with ``password``. Defaults to None.
		password (str, optional): Password to log in with; used in conjunction with ``username``. Defaults to None.
		key (str, optional): Feature Key to log in as. Defaults to "AppscanEnterpriseUser".
		apikey (str, optional): API Key to log in with; used in conjunction with ``secretkey``. Defaults to None.
		secretkey (str, optional): API Secret to log in with; used in conjunction with ``apikey``. Defaults to None.
		verify (bool, optional): Certificate path for requests. Defaults to False.
	
	Returns:
		:class:`pyappscan.pyappscan.PyAppscan`
	"""
	return PyAppscan(url, username, password, key, apikey, secretkey, verify)



from .helper import Helper
import re
from typing import List, Dict, Union, Tuple, Any

class Query(Helper):
	"""Do not call this directly; use :func:`pyappscan.create_engine`"""
	def __init__(self, baseurl: str, username: str, password: str, key: str, apikey: str, secretkey: str, verify: bool):
		super(Query, self).__init__(baseurl=baseurl, username=username, password=password, key=key, apikey=apikey, secretkey=secretkey, verify=verify)

	def getapikey(self):
		'''Gets API key and API secret from appscan; requires :attr:`username` and :attr:`password` to work'''
		r = self.login()
		result = r.get(self.baseurl+'account/getapikey', verify=self.verify)
		return self.returnobject(result)

	def templatequery(self, name: str = None, category: str = None) -> Union[Dict[str, str], List[dict]]:
		'''Queries template endpoint.
		
		Omit category if using element ID as name argument to query element directly.

		Args:
			name (str, optional): Key to search for. Defaults to ``None``.
			category (str, optional): Value to search for. Defaults to ``None``.

		References:
			:meth:`pyappscan.helper.Helper.queryhelper`
		'''
		return self.queryhelper('templates', name, category)
	
	def folderquery(self, name: str = None, category: str = None) -> Union[Dict[str, str], List[dict]]:
		'''Queries folder endpoint

		Omit category if using element ID as name argument to query element directly.

		Args:
			name (str, optional): Key to search for. Defaults to ``None``.
			category (str, optional): Value to search for. Defaults to ``None``.

		References:
			:meth:`pyappscan.helper.Helper.queryhelper`
		'''
		return self.queryhelper('folders', name, category)

	def testpolicyquery(self, name: str = None, category: str = None) -> Union[Dict[str, str], List[dict]]:
		'''Queries test policy endpoint

		Omit category if using element ID as name argument to query element directly.

		Args:
			name (str, optional): Key to search for. Defaults to ``None``.
			category (str, optional): Value to search for. Defaults to ``None``.

		References:
			:meth:`pyappscan.helper.Helper.queryhelper`
		'''
		return self.queryhelper('testpolicies', name, category)
	
	def applicationquery(self, name: str = None, category: str = None) -> Union[Dict[str, str], List[dict]]:
		'''Queries application endpoint

		Omit category if using element ID as name argument to query element directly.

		Args:
			name (str, optional): Key to search for. Defaults to ``None``.
			category (str, optional): Value to search for. Defaults to ``None``.

		References:
			:meth:`pyappscan.helper.Helper.queryhelper`
		'''
		return self.queryhelper('applications', name, category)

	def scannerquery(self, name: str = None, category: str = None) -> Union[Dict[str, str], List[dict]]:
		'''Queries scanner endpoint

		Omit category if using element ID as name argument to query element directly.

		Args:
			name (str, optional): Key to search for. Defaults to ``None``.
			category (str, optional): Value to search for. Defaults to ``None``.

		References:
			:meth:`pyappscan.helper.Helper.queryhelper`
		'''
		r = self.login()
		result = r.get(self.baseurl+'scanners?includeUnregisteredScanners=true', verify=self.verify)
		if name == None:
			return self.returnobject(result)
		else:
			i = self.searchdict(result.json(), name, category)
			return i
	
	def userquery(self, name: str = None, category: str = None) -> Union[Dict[str, str], List[dict]]:
		'''Queries user endpoint

		Omit category if using element ID as name argument to query element directly.

		Args:
			name (str, optional): Key to search for. Defaults to ``None``.
			category (str, optional): Value to search for. Defaults to ``None``.

		References:
			:meth:`pyappscan.helper.Helper.queryhelper`
		'''
		return self.queryhelper('consoleusers', name, category)

	def usertypequery(self, name: str = None, category: str = None) -> Union[Dict[str, str], List[dict]]:
		'''Queries usertype endpoint

		Omit category if using element ID as name argument to query element directly.

		Args:
			name (str, optional): Key to search for. Defaults to ``None``.
			category (str, optional): Value to search for. Defaults to ``None``.

		References:
			:meth:`pyappscan.helper.Helper.queryhelper`
		'''
		return self.queryhelper('usertypes', name, category)

	def issuetypes(self, name: str = None, category: str = None) -> Union[Dict[str, str], List[dict]]:
		'''Queries issue type endpoint

		Omit category if using element ID as name argument to query element directly.

		Args:
			name (str, optional): Key to search for. Defaults to ``None``.
			category (str, optional): Value to search for. Defaults to ``None``.

		References:
			:meth:`pyappscan.helper.Helper.queryhelper`
		'''
		return self.queryhelper('issuetypes', name, category)

	def currentscans(self, name: str = None, category: str = None) -> Union[Dict[str, str], List[dict]]:
		'''Queries scan queue/management endpoint

		Omit category if using element ID as name argument to query element directly.

		Args:
			name (str, optional): Key to search for. Defaults to ``None``.
			category (str, optional): Value to search for. Defaults to ``None``.

		References:
			:meth:`pyappscan.helper.Helper.queryhelper`
		'''
		return self.queryhelper('scansmanagement', name, category)
	
	def jobsquery(self, name: str = None, category: str = None) -> Union[Dict[str, str], List[dict]]:
		'''Queries jobs endpoint for a given folder or job

		Returns all DAST-configured jobs in a folder. Given a ``name`` as a job 
		ID and no ``category``, this will return the entry for the given 
		job ID. If given no ``name``, this will return all jobs in the 
		console. If given a ``name`` and a ``category``, this will return the job 
		that matches the category in a manner similar to queryhelper.

		Args:
			name (str, optional): Key to search for. Defaults to ``None``.
			category (str, optional): Value to search for. Defaults to ``None``.
		
		Returns:
			dict OR list: If you have both or neither, a :class:`dict` is returned, 
			otherwise :class:`list` is returned
		
		References:
			:meth:`folderjobsquery`
		'''
		r = self.login()
		jobslist = []
		if name and not category:
			result = r.get(self.baseurl + '/jobs/' + str(name))
			return self.returnobject(result)
		folderlist = self.folderquery()
		for folder in folderlist:
			if name == None:
				jobquery = self.folderjobsquery(foldername=folder['folderName'])
				jobslist.append(jobquery)
			else:
				jobquery = self.folderjobsquery(foldername=folder['folderName'], jobname=name)
				if jobquery:
					return jobquery
				else:
					continue
		return jobslist

	def folderjobsquery(self, foldername: Union[str, int], jobname: Union[str, int] = None) -> Union[List[dict], Dict[str, str]]:
		"""Queries jobs endpoint of folder path

		Args:
			foldername (Union[str, int]): ID or name of folder to query; if :class:`str` 
			grabs ID for you
			jobname (Union[str, int], optional): Job ID or name; if :class:`str` grabs ID 
			for you. Defaults to ``None``.

		Returns:
			Union[List[dict], Dict[str, str]]: Returns :class:`list` if only ``foldername`` is given, 
			returns :class:`dict` otherwise
		"""
		r = self.login()
		try:
			if bool(int(foldername)) == True: 
				folderid = int(foldername)
		except ValueError:
			folderid = self.folderquery(foldername, 'folderName')['folderId']
		result = r.get(self.baseurl+'folders/'+str(folderid)+'/jobs', verify=self.verify)
		if jobname == None:
			return self.returnobject(result)
		else:
			try:
				if bool(int(jobname)) == True:
					return self.searchlist(result.json(), jobname, 'id')
			except ValueError:
				return self.searchlist(result.json(), jobname, 'name')

	def appissuequery(self, appname: Union[str, int] = None, severity: str = 'High', sortBy: str = 'Severity', itemcount: int = 100, includeinfoseverity: bool = False) -> List[dict]:
		"""Queries issues endpoint using application path

		Args:
			appname (str OR int, optional): Application ID or name. If :class:`str` 
			grabs app ID for you. Defaults to ``None``.
			severity (str, optional): Severity of issues to include. Defaults to 'High'.
			sortBy (str, optional): Sorting category. Defaults to 'Severity'.
			itemcount (int, optional): Amount of issues to return. Defaults to 100.
			includeinfoseverity (bool, optional): Include issues with informational 
			severity. Defaults to ``False``.

		Returns:
			list: list of issue dictionaries
		"""
		r = self.login()
		if appname == None:
			jsonlist = []
			appjson = self.applicationquery()
			applist = []
			for node in appjson:
				applist.append(node['name'])
			for node in applist:
				url = self.baseurl+'issues?columns=Severity,Location,Issue Type,Element&query=Application Name='+node
				if severity == 'Medium' or severity == 'medium': sev=['Medium', 'Low']
				elif severity == 'Low' or severity == 'low': sev=['Low']
				else: sev=['High', 'Medium', 'Low']
				if includeinfoseverity == True:
					sev.append('Information')
				for entry in sev:
					url+=',Severity='+entry
				url+='&sortBy=-'+sortBy+'&compactResponse=false'
				result = r.get(url, verify=self.verify)
				jsonlist.append(result.json())
			return jsonlist
		else:
			try:
				if bool(int(appname)) == True:
					appname = self.applicationquery(appname, 'id')['name']
			except ValueError:
				pass
			url = self.baseurl+'issues?columns=Severity,Location,Issue Type,Element&query=Application Name='+appname
			if severity == 'Medium' or severity == 'medium': sev=['Medium', 'Low']
			elif severity == 'Low' or severity == 'low': sev=['Low']
			else: sev=['High', 'Medium', 'Low']
			if includeinfoseverity == True:
				sev.append('Information')
			for entry in sev:
				url+=',Severity='+entry
			url+='&sortBy=-'+sortBy+'&compactResponse=false'
			header = {'Range':'items=0-{}'.format(itemcount)}
			r.headers.update(header)
			result = r.get(url, verify=self.verify)
			return self.returnobject(result)
	
	def dastquery(self, jobid: int) -> Dict[str, str]:
		"""Gets the DAST configuration of the job

		Args:
			jobid (int): ID of job

		Returns:
			dict: DAST configuration
		"""
		r = self.login()
		result = r.get(self.baseurl+'jobs/'+str(jobid)+'/dastconfig', verify=self.verify)
		return self.returnobject(result)

	def permissionsquery(self) -> Dict[str, str]:
		"""Get permissions of the current user

		Returns:
			dict: permissions dictionary
		"""
		r = self.login()
		result = r.get(self.baseurl+'currentuser_v2', verify=self.verify)
		return self.returnobject(result)
	
	def schedulequery(self, jobid: int) -> Dict[str, str]:
		"""Get the current schedule of the job

		Args:
			jobid (int): ID of the job

		Returns:
			dict: Dictionary of the schedule of the scan
		"""
		r = self.login()
		result = r.get(self.baseurl+'jobs/schedulescan/'+str(jobid)+'?responseFormat=json', verify=self.verify)
		return self.returnobject(result)

	def reportquery(self, reportid: int, issues: bool = False) -> bytes:
		"""Report information from XML reports endpoint; please use :meth:`reportjsonquery` instead

		Note:
			This only exists for backwards compatibility; please use :meth:`reportjsonquery` instead

		Args:
			reportid (int): Report ID
			issues (bool, optional): Whether to get issue report (default is summary report). Defaults to False.

		Returns:
			bytes: HTML/XML of the endpoint

		References:
			:meth:`reportjsonquery`
		"""
		r = self.login()
		if issues == False:
			result = r.get(self.scanbaseurl+'services/reports/'+str(reportid), verify=self.verify)
		else:
			result = r.get(self.scanbaseurl+'services/reports/'+str(reportid)+'/issues', verify=self.verify)
		return self.returnhtml(result)

	def reportjsonquery(self, reportid: int, issues: bool = False, summary: bool = False) -> Union[List[dict], Dict[str, Any]]:
		"""Get report data in JSON format

		By default you hit the `reports/{reportid}/data` endpoint, from which you will 
		get a detailed data dictionary in which the issues list usually at ``wf-security-issues.issue``. 
		If you set ``issues`` to :class:`True`, you get a detailed issue dictionary
		which has the issue list at ``issues.security-issue``. If you set ``summary`` 
		to :class:`True` you will get an issue summary by severity and status.

		Args:
			reportid (int): Report ID
			issues (bool, optional): Get detailed issues list. Defaults to False.
			summary (bool, optional): Get summary of issues. Defaults to False.

		Returns:
			Union[List[dict], Dict[str, Any]]: This method tries to filter out the nonsense and leave 
			only the issue lists, but for some reason sometimes the base dictionary
			structure changes, so sometimes we just return the dictionary.
		"""
		r = self.login()
		result = ''
		try:
			if bool(int(reportid)) == True:
				reportid = int(reportid)
		except:
			reportlist = self.folderitemquery(folderitemid=reportid, reports=True)['id']
			reportid = int(reportlist)
		if issues == False and summary == True:
			result = r.get(self.baseurl+'reports/' + str(reportid), verify=self.verify)
		elif issues == True and summary == False:
			url = self.baseurl+'reports/' + str(reportid) + '/issues'
			result = r.get(url, verify=self.verify)
			json_result = self.returnobject(result)
			if json_result['issues']['TotalPages'] != 1:
				for page in range(2, json_result['issues']['TotalPages']):
					paged_result = r.get(url+'/pages/'+str(page), verify=self.verify)
					json_paged_result = self.returnobject(paged_result)
					json_result['issues']['security-issue'] = json_result['issues']['security-issue'] + json_paged_result['issues']['security-issue']
			return json_result
		else:
			url = self.baseurl+'reports/' + str(reportid) + '/data'
			result = r.get(url, verify=self.verify)
			json_result = self.returnobject(result)
			if json_result['wf-security-issues']['TotalPages'] != 1:
				for page in range(2, json_result['wf-security-issues']['TotalPages']):
					paged_result = r.get(url+'/pages/'+page, verify=self.verify)
					json_paged_result = self.returnobject(paged_result)
					json_result['wf-security-issues']['security-issue'] = json_result['wf-security-issues']['security-issue'] + json_paged_result['wf-security-issues']['security-issue']
			return json_result
		return self.returnobject(result)

	def issuequery(self, reportid: int = None, appid: Union[str, int] = None, issueid: int = None, zipfile: bool = False, issuecount: int = 100) -> Union[List[dict], Dict[str, Any]]:
		"""Grabs issues from either a report or an app, or grabs issue details

		Provided a ``reportid``, this will use :meth:`reportquery` to grab issues, or if 
		provided an ``appid``, this will use :meth:`appissuequery` to grab issues, or if 
		provided an ``issueid`` in concurrence with either one will return that 
		specific isssue from the requested app or report.

		Note:
			This method is deprecated due to being discouraged when the ability 
			to call :meth:`reportjsonquery` or :meth:`appissuequery` is feasible, 
			and for using :meth:`reportquery` instead of :meth:`reportjsonquery`.

		Args:
			reportid (int, optional): Report ID. Defaults to None.
			appid (Union[str, int], optional): App ID; if string, app ID will be grabbed. Defaults to None.
			issueid (int, optional): Issue ID. Defaults to None.
			zipfile (bool, optional): Should the app issues be returned in ZIP format. 
				Note that this cannot be :class:`True` while ``reportid`` is given. Defaults to False.
			issuecount (int, optional): Issue count to be returned. Defaults to 100.

		Raises:
			TypeError: Raised if ``zipfile`` is :class:`True` while ``reportid`` is given.

		Returns:
			Union[List[dict], Dict[str, Any]]: Returns issue :class:`list` if ``issueid`` is not provided, and
			:class:`dict` if it is provided.

		References:
			* :meth:`appissuequery`
			* :meth:`reportjsonquery`
			* :meth:`reportquery`
		"""		
		r = self.login()
		if reportid != None:
			if zipfile != False:
				raise TypeError("You can't return a report issue in zip format")
			#if reportid is given but issue id isnt
			if issueid == None:
				liss = []
				for entry in re.finditer('\\<id\\>[\\d]+\\<\\/id\\>', self.reportquery(reportid, issues=True).decode('utf-8')):
					ent = entry.group()
					ent = int(ent[ent.find('>')+1:ent.find('<', 2)])
					if ent != reportid:
						if ent > 1000:
							liss.append(self.issuequery(reportid, issueid=ent).decode('utf-8'))
				return liss
			#if reportid is given and issue id is as well
			else:
				result = r.get(self.scanbaseurl+'services/reports/'+str(reportid)+'/issues/'+str(issueid), verify=self.verify)
				return self.returnhtml(result)
		elif appid != None:
			try:
				if bool(int(appid)) == True:
					appid = int(appid)
			except ValueError:
				appid = self.applicationquery(appid, 'name')['id']
			#if appid is given but issueid isnt
			if issueid == None:
				return self.appissuequery(appname=appid, itemcount=issuecount)
			#if appid is given and so is issue id
			else:
				if zipfile == False:
					result = r.get(self.baseurl+'issues/'+str(issueid)+'/application/'+str(appid), verify=self.verify)
				else:
					result = r.get(self.baseurl+'issues/details_v2?appId='+str(appid)+'&ids=["'+str(issueid)+'"]', verify=self.verify)
			return self.returnobject(result)
		else: return ValueError("Please have either the reportid or the appid in your arguments")

	def aboutissue(self, appid: Union[int, str], issueid: int) -> bytes:
		"""Gets HTML document about given issue ID

		Args:
			appid (Union[int, str]): Application ID; if string, app ID will be grabbed.
			issueid (int): Issue ID

		Returns:
			bytes: HTML encoded in UTF-8 format
		"""
		r = self.login()
		try:
			if bool(int(appid)) == True:
				appid = int(appid)
		except ValueError:
			appid = self.applicationquery(appid, 'name')['id']
		result = r.get(self.baseurl+'issues/'+str(issueid)+'/application/'+str(appid)+'/aboutthisissue', verify=self.verify)
		return self.returnhtml(result)

	def issuetypequery(self, issuetypeid: int = None, lookuptype: str = None, lookupname: str = None) -> Union[bytes, Dict[str, Any]]:
		"""Gets data about issue type depending on information given

		Args:
			issuetypeid (int, optional): Issuetype ID. If :class:`None`, ``lookupname`` is required. Defaults to None.
			lookuptype (str, optional): Type of information to look up; defaults to None. Accepted values are:
				
				* ``advisory``
				* ``fix``

			lookupname (str, optional): Lookup string identifier; usually ``wf-*``. If 
				:class:`None`, lookup name associated with issue type is looked up and searched. 
				Defaults to None.

		Returns:
			Union[bytes, Dict[str, Any]]: Returns :class:`bytes` type HTML information for a lookupname or :class:`dict` when no ``lookuptype`` and ``lookupname`` is given.
		"""
		r = self.login()
		if issuetypeid != None:
			try:
				if bool(int(issuetypeid)) == True:
					issuetypeid = int(issuetypeid)
			except ValueError:
				issuetypeid = self.issuetypes(issuetypeid, category="name")['id']
		if lookuptype == None:
			result = r.get(self.baseurl+'issuetypes/'+str(issuetypeid), verify=self.verify)
			return self.returnobject(result)
		elif lookuptype == 'advisory' or lookuptype == 'Advisory':
			if lookupname != None:
				result = r.get(self.baseurl+'issuetypes/'+str(lookupname)+'/advisory', verify=self.verify)
			else:
				if issuetypeid == 0:
					lookupname = lookupname
				else:
					lookupname = self.issuetypequery(issuetypeid)['issueTypeLookup']
				result = r.get(self.baseurl+'issuetypes/'+str(lookupname)+'/advisory', verify=self.verify)
			return self.returnhtml(result)
		elif lookuptype == 'fix' or lookuptype == 'Fix':
			if lookupname != None:
				result = r.get(self.baseurl+'issuetypes/'+str(lookupname)+'/fixrecommendation', verify=self.verify)
			else:
				if issuetypeid == 0:
					lookupname = lookupname
				else:
					lookupname = self.issuetypequery(issuetypeid)['issueTypeLookup']
				result = r.get(self.baseurl+'issuetypes/'+str(lookupname)+'/advisory', verify=self.verify)
			return self.returnhtml(result)

	def applicationattributequery(self, appid: Union[int, str] = None) -> Dict[str, List[dict]]:
		"""Gets attributes of application

		Args:
			appid (int OR str, optional): App ID; if string, the app ID is grabbed. Defaults to None.

		Returns:
			Dict[str, List[dict]]: dictionary of attributes with attribute list inside.
		"""
		r = self.login()
		try:
			if bool(int(appid)) == True:
				result = r.get(self.baseurl+'applications/'+str(appid), verify=self.verify)
				return self.returnobject(result)
		except TypeError:
			if type(appid) == str:
				appid = self.applicationquery(appid, 'name')['id']
				result = r.get(self.baseurl+'applications/'+str(appid), verify=self.verify)
				return self.returnobject(result)
			else:
				result = r.get(self.baseurl+'appattributedefinitions', verify=self.verify)
				return self.returnobject(result)

	def issueattributequery(self) -> Dict[str, List[dict]]:
		"""Gets issue attributes JSON schema

		Returns:
			Dict[str, List[dict]]: dictionary of attirbutes
		"""
		r = self.login()
		result = r.get(self.baseurl + 'issueattributedefinitions', verify=self.verify)
		return self.returnobject(result)

	def folderitemsquery(self, folderid: Union[int, str], keyword: str = None, name: str = None, category: str = None) -> Dict[str, Any]:
		"""Get all items in a folder, DAST-compliant or not.

		Args:
			folderid (int OR str): Folder ID; if string, folder ID will be grabbed.
			keyword (str, optional): Keyword if you know the type of folder items you're 
			looking for. Defaults to None. Accepted values:
				``report-pack``
				``content-scan-job``
			name (str, optional): Value to search for in ``category``. Defaults to None.
			category (str, optional): Key to search for ``name``. Defaults to None.

		Returns:
			dict: Report pack, scan jobs, or both, depending on the keyword or lack thereof

		References:
			:meth:`pyappscan.helper.Helper.queryhelper`
		"""
		r = self.login()
		try:
			if bool(int(folderid)) == True:
				folderid = int(folderid)
		except ValueError:
			if type(folderid) == str:
				folderid = int(self.folderquery(name=folderid, category='folderName')['folderId'])
		result = r.get(self.baseurl +'folders/'+str(folderid)+'/folderitems', verify=self.verify)
		try:
			jsondict = result.json()['folder-items']
			if keyword != None:
				jsondict = jsondict[keyword]
			if name != None and category != None:
				if keyword == None:
					return self.searchdict(jsondict, name, category)
				else:
					return self.searchlist(jsondict, name, category)
			else:
				return jsondict
		except:
			return self.returnobject(result)

	def folderitemquery(self, folderitemid: int, options: bool = False, reports: bool = False, scanlog: bool = False, statistics: bool = False):
		"""Gets a given folder item's requested information

		``options`` returns the options associated with the folder item
		``reports`` returns report IDs and info associated with the folder item
		``scanlog`` returns the scan log of the folder item
		``statistics`` returns the statistics of the folder item

		Args:
			folderitemid (int): Folder item ID
			options (bool, optional): returns the options associated with the folder item. Defaults to False.
			reports (bool, optional): returns report IDs and info associated with the folder item. Defaults to False.
			scanlog (bool, optional): returns the scan log of the folder item. Defaults to False.
			statistics (bool, optional): returns the statistics of the folder item. Defaults to False.

		Raises:
			ValueError: Raised if you provide a string for ``folderitemid``

		Returns:
			dict: Returns information requested
		"""
		r = self.login()
		# decide if folderitemid is an id or the name of the folderitem
		try:
			# if it is, all good
			if bool(int(folderitemid)) == True:
				folderitemid = int(folderitemid)
		except ValueError:
			if type(folderitemid) == str:
				# if it's not, get all folders and iterate through them for folderitems until we find the name we're looking for
				folderjson = self.folderquery()
				for item in folderjson:
					potential_item_json = self.folderitemsquery(folderid=item['folderId'], name=folderitemid, category='name')
					if potential_item_json == None:
						continue
					else:
						folderitemid = int(potential_item_json['id'])
				else:
					raise ValueError("Not a valid folderitem name")
		result = ''
		if options and not (reports or scanlog or statistics):
			result = r.get(self.baseurl+'folderitems/'+str(folderitemid)+'/options', verify=self.verify)
		elif reports and not (options or scanlog or statistics):
			result = r.get(self.baseurl+'folderitems/'+str(folderitemid)+'/reports', verify=self.verify).json()['reports']['report']
			result = [rep for rep in result if rep['name'] == 'Security Issues'][0]
		elif scanlog and not (options or reports or statistics):
			result = r.get(self.baseurl+'folderitems/'+str(folderitemid)+'/scanlog', verify=self.verify)
		elif statistics and not (options or reports or scanlog):
			result = r.get(self.baseurl+'folderitems/'+str(folderitemid)+'/statistics', verify=self.verify)
		else:
			result = r.get(self.baseurl+'folderitems/'+str(folderitemid), verify=self.verify)
		if type(result) == dict:
			return result
		return self.returnobject(result)

	def servergroupquery(self) -> Dict[str, dict]:
		r = self.login()
		result = r.get(self.baseurl + 'servergroups?responseFormat=json', verify=self.verify)
		return self.returnobject(result)

	def dashboardquery(self, businessunit: str) -> bytes:
		"""Queries dashboard

		Note:
			This is method is still in beta

		Args:
			businessunit (str): Business unit associated with the dashboard

		Returns:
			bytes: HTML of dashboard
		"""
		r = self.login()
		result = r.get(self.baseurl+'dashboard?query=Business Unit='+businessunit, verify=self.verify)
		return self.returnhtml(result)

	def buildfoldertree(self) -> str:
		"""Used to print out the entire tree of the folder structure of appscan

		Returns:
			str: formatted tree string
		"""
		folderjson = self.folderquery()
		folders = {}
		folderdict = {}
		for i in folderjson:
			folderid = i['folderId']
			foldername = i['folderName']
			if folderid == 1:
				folderdict[folders[1]] = {}
			dirlist = i['folderPath'].split('/')
			dirlist = dirlist[:-1]
			for path in dirlist[:-1]:
				pointer = pointer[path]
			pointer[foldername] = {}	
		return self.dictprint(folderdict)	



import requests
from os.path import isfile
import logging
from .exceptions import LogoutError, LoginError, AttrError
from datetime import timedelta
import os
from typing import List, Dict, Union, Any, Tuple

try:
	from json import JSONDecodeError
except ImportError:
	JSONDecodeError = ValueError

class Helper(object):
	"""Helper class which is inherited by all other classes; do not use this class directly, use :func:`pyappscan.create_engine`"""
	def __init__(self, baseurl, username, password, key, apikey, secretkey, verify):
		i = baseurl.find(':')
		scanbaseurl1 = baseurl[:baseurl.find(':', i+1)]
		i = baseurl.find('/ase' or '/ASE')
		scanbaseurl2 = baseurl[i:i+4]+'/'
		self.scanbaseurl = scanbaseurl1+scanbaseurl2
		if baseurl[-1] == '/':
			self.baseurl = baseurl
		else:
			self.baseurl = baseurl+'/'
		if username:
			self.username = username
			self.password = password
			if key==None:
				self.key = 'AppscanEnterpriseUser'
			else:
				self.key = key
		self.apikey = apikey
		self.secretkey = secretkey
		self.verify = verify

		# Logger config
		logger = logging.getLogger(__name__)
		logging_level = logging.ERROR
		logging.captureWarnings(True)
		logger.setLevel(logging_level)
		handler = logging.StreamHandler()
		handler.setLevel(logging_level)
		if logger.handlers:
			logger.handlers = []
		formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
		handler.setFormatter(formatter)
		logger.addHandler(handler)
		self.logger = logger

	def queryhelper(self, url_ext: str, name: str = None, category: str = None) -> Union[Dict[str, str], List[dict]]:
		"""Decides how to query endpoint data

		If name is given but category is not,
		returns query of ``url_ext`` with ``name`` appended for direct item query. If
		name is not given, it returns the query of ``url_ext``. If name and
		category are given, it returns the query of ``url_ext`` after searching for
		value ``name`` in key ``category``.

		Args:
			url_ext (str): Extension of :attr:`baseurl` to query
			name (str, optional): Value to search for; either appended to ``url_ext`` 
				or searched for in key ``category``. Defaults to None.
			category (str, optional): Key to search ``name`` in. Defaults to None.

		Returns:
			Union[Dict[str, str], List[str]]: either returns a dictionary or a list; will return dictionary 
				if ``name`` is given or if ``name`` and ``category`` are given.
		"""		
		r = self.login()
		if name != None and category == None:
			result = r.get(self.baseurl + url_ext + '/' + str(name), verify=self.verify)
			return self.returnobject(result)
		result = r.get(self.baseurl+url_ext, verify=self.verify)
		if name == None:
			return self.returnobject(result)
		else:
			return self.searchlist(result.json(), name, category)

	def searchlist(self, liss: List[dict], name: str, category: str) -> Union[Dict[str, Any], None]:
		"""Searches a python list of dictionaries for dict containing value ``name`` and key ``category``.

		Args:
			liss (List[dict]): List of dictionaries to be searched
			name (str): Value to be searched for
			category (str): Key to search ``name`` in

		Raises:
			TypeError: Raised when ``liss`` is not type :obj:`list`

		Returns:
			Union[Dict[str, Any], None]: returns dictionary if found, :obj:`None` if not found.
		"""
		if type(liss) != list:
			raise TypeError("Must be a list")
		for node in liss:
			if name == node[category]:
				return node
			elif str(name) == node[category]:
				return node
		return None
	
	def searchdict(self, dic: Dict[str, list], name: str, category: str) -> Union[Dict[str, Any], None]:
		"""Searches a dictionary of lists for a value

		Args:
			dic (Dict[str, list]): Dictionary of lists
			name (str): Value to be searched for
			category (str): Key to search ``name`` in

		Raises:
			TypeError: Raised when ``dic`` is not type :obj:`dict`

		Returns:
			Union[Dict[str, Any], None]: returns ``dict`` when found, :obj:`None` if not found
		"""
		if type(dic) != dict:
			raise TypeError("Must be a dictionary")
		for value in dic.values():
			res = self.searchlist(value, name, category)
			if res == None:
				continue
			else:
				return res
		return None

	def version(self) -> Dict[str, str]:
		"""Returns version of appscan

		Returns:
			Dict[str, str]: Version dictionary
		"""
		r = requests.session()
		result = r.get(self.baseurl+'version', verify=self.verify)
		return self.returnobject(result)

	def login(self) -> requests.Session:
		"""Logs into the server; you shouldn't have to call this yourself

		Raises:
			LoginError: Raised if Login was unsuccessful

		Returns:
			:class:`requests.Session`: Requests session with login cookies
		"""
		r = requests.session()
		if hasattr(self, 'username'):
			dic = {'userId':self.username,'password':self.password,'featureKey':self.key}
			result = r.post(self.baseurl+'login', json=dic, verify=self.verify)
		if self.apikey:
			dic = {'keyId':self.apikey,'keySecret':self.secretkey}
			result = r.post(self.baseurl+'keylogin/apikeylogin', json=dic, verify=self.verify)
		try:
			self.logger.debug("message=\"Login result json\", output=\"{}\"".format(result.json()))
			headers = {'asc_xsrf_token':result.json()['sessionId']}
		except JSONDecodeError:
			raise LoginError()
		r.headers.update(headers)
		return r
	
	def returnobject(self, obj: requests.Response) -> Union[dict, requests.Response]:
		"""Helper function to return json or log and return errors with the request

		Args:
			obj (:class:`requests.Response`): a requests response object from appscan

		Returns:
			Union[dict, :class:`requests.Response`]: if json can be coerced, return JSON, else return object back
		"""
		if obj.status_code == 200 or obj.status_code == 201 or obj.status_code == 204:
			try:
				return obj.json()
			except:
				return obj
		else:
			self.logger.error(obj.content)
			obj.raise_for_status()

	def returnhtml(self, obj: requests.Response) -> bytes:
		"""Helper function to return HTML content from a response object

		Args:
			obj (:class:`requests.Response`): a requests response object from appscan

		Returns:
			bytes: Returns binary unicode content that must be decoded
		"""
		if obj.status_code == 200 or obj.status_code == 201 or obj.status_code == 204:
			return obj.content
		else:
			self.logger.error(obj.content)
			obj.raise_for_status()
	
	def dicttolistofdicts(self, dic: Dict[str, Any]) -> List[Dict[str, Any]]:
		"""Converts a dictionary to a list of dictionaries

		Args:
			dic (Dict[str, Any]): dictionary to be converted

		Returns:
			List[Dict[str, Any]]: list of dictionaries
		"""
		liss = []
		if type(dic['URL']) != list:
			for k,v in dic.items():
				if type(v) != str:
					v = str(v)
				newdic={'name' : k, 'value' : [v]}
				liss.append(newdic)
		else:
			for k,v in dic.items():
				if type(v) != str:
					v = str(v)
				newdic={'name' : k , 'value' : v}
				liss.append(newdic)
		return liss

	def logout(self, session: requests.Session) -> requests.Response:
		"""Logs out of appscan

		Args:
			session (:class:`requests.Session`): a requests session object that has been logged into appscan with :meth:`login`

		Raises:
			LogoutError: Raised when logout failed

		Returns:
			:class:`requests.Response`: Requests response object after logging out
		"""
		if type(session) != requests.sessions.Session:
			raise LogoutError()
		result = session.get(self.baseurl+'logout', verify=self.verify)
		return result

	def dictprint(self, dic: Dict[Any, Any], string: str = '', tab: int = 0) -> str:
		for k,v in dic.items():
			for _ in range(tab):
				string+='|\t'
			string+='| '+k+'\n'
			if type(v) == dict:
				if not v:
					pass
				else:
					string = self.dictprint(v, string=string, tab=tab+1)
		return string

	def attributecheck(self, attributedict):
		"""Checks attributes for :meth:`pyappscan.pyappscan.PyAppscan.createapplication`

		Args:
			attributedict (dict): attribute dictionary

		Raises:
			AttrError: Raised if attribute not present
		
		References:
			* :meth:`pyappscan.pyappscan.PyAppscan.createapplication`
			* :meth:`pyappscan.pyappscan.PyAppscan.updateapplication`
		"""
		if len(attributedict.keys()) < 6:
			raise AttrError('len')
		if 'Name' not in attributedict.keys():
			raise AttrError('Name')
		if 'URL' not in attributedict.keys():
			raise AttrError('URL')
		if 'Hosts' not in attributedict.keys():
			raise AttrError('Hosts')
		if 'Development Contact' not in attributedict.keys():
			raise AttrError('Development Contact')
		if 'Business Unit' not in attributedict.keys():
			raise AttrError('Business Unit')
		if 'Description' not in attributedict.keys():
			raise AttrError('Description')

	def deltaToString(self, delta: timedelta) -> Tuple[str, str]:
		"""Converts a python :class:`datetime.timedelta` object to an appscan scheduling string

		Args:
			delta (:class:`datetime.timedelta`): python timedelta object

		Raises:
			ValueError: raised when delta is not timedelta

		Returns:
			Tuple[str, str]: tuple of the datestring, and the type of datestring respectively
		"""
		string = ''
		type_ = ''
		if type(delta) != timedelta:
			raise ValueError("delta must be type timedelta")
		schedule_months = delta.days / 30
		schedule_weeks = delta.days / 7
		schedule_month_days = delta.days % 30
		schedule_week_days = delta.days % 7
		if schedule_week_days == 0:
			schedule_week_days = 1
		if schedule_month_days == 0:
			schedule_month_days = 1
		if schedule_months == 0:
			if schedule_weeks == 0:
				string = "{};5:00PM".format(schedule_week_days)
				type_ = 'daily'
			else:
				string = "{};{};5:00PM".format(schedule_weeks, schedule_week_days)
				type_ = 'weekly'
		else:
			string = "{};{};5:00PM".format(schedule_months, schedule_month_days)
			type_ = 'monthly'
		return string, type_



class LogoutError(Exception):
	"""Logout failed"""
	def __init__(self):
		Exception.__init__(self, "LOGOUT FAILED: Object passed is not a session object.")

class LoginError(Exception):
	"""Login failed"""
	def __init__(self):
		Exception.__init__(self, "LOGIN FAILED: Credentials invalid. Please check your username, password, and featureKey to make sure they are all correct.")

class UserTypeError(Exception):
	"""Bad Usertype"""
	def __init__(self):
		Exception.__init__(self, "USER TYPE INVALID: please specify a user type with an integer of a known user type, or with the name of a known user type. If you aren't sure which usertypeID does what, call\nfor i in PyAppscan.usertypequery():\n\tfor k,v in i.items():\n\t\tprint(k,v)")

class DataCodeError(Exception):
	"""Bad Data Code"""
	def __init__(self):
		Exception.__init__(self, 'Please enter a valid data code. Valid data codes are:\n\turl\n\tusername\n\tpassword\n\tmethod\n\theader\n\tlockout\n\tadditionaldomains\n\texclusions')

class AttrError(Exception):
	"""Bad attribute in app attribute dictionary"""
	def __init__(self, *args):
		if args[0] == 'len':
			Exception.__init__(self, 'You must have the required amount of dictionary entries in your attributedict (6), and they must be the same as the ones in the documentation.')
		else:
			Exception.__init__(self, 'You must have %s in your attributedict; it seems to be missing, or it is not spelled correctly. Please check the documentation for spelling, as it is case sensetive' % args[0])



from setuptools import setup, find_packages
import os

def here(*path):
    return os.path.join(os.path.dirname(__file__), *path)

def get_file_contents(filename):
    with open(here(filename)) as fp:
        return fp.read()

long_description = get_file_contents("README.md")

# This is a quick and dirty way to include everything from
# requirements.txt as package dependencies.
install_requires = get_file_contents('requirements.txt').split()

packageinfo = {}
exec(open('pyappscan/__packageinfo__.py').read(), packageinfo)

setup(
	name='pyappscan',
	description='Python REST API Wrapper for IBM/HCL Appscan, a tool which helps scan and report on website and mobile app vulnerabilities.',
	long_description=long_description,
	long_description_content_type='text/markdown',
	url=packageinfo['giturl'],
	author=packageinfo['author'],
	author_email=packageinfo['author_email'],
	version=packageinfo['version'],
	packages=find_packages(exclude=['tests']),
	include_package_data=True,
	install_requires=install_requires,
	python_requires=">=3.5.0",
	zip_safe=False
)



from .exceptions import UserTypeError, DataCodeError
from time import time
from datetime import datetime
from .query import Query
from typing import Dict, Union, Any

class PyAppscan(Query):
	"""Main class that is returned when called from :func:`pyappscan.create_engine`

	Note:
		You may be wondering why there is no ``deletejob`` method. This is because you
		cannot directly delete jobs via the API, you have to delete the folder item.
		This can be done using the :meth:`deletefolderitem` method, during which you
		should also delete the associated report, as not deleting it causes issues with
		Appscan internals.

	Args:
		baseurl (str): Base URL for appscan
		username (str, optional): Username to log in with; used in conjunction with ``password``. Defaults to None.
		password (str, optional): Password to log in with; used in conjunction with ``username``. Defaults to None.
		key (str, optional): Feature Key to log in as. Defaults to "AppscanEnterpriseUser".
		apikey (str, optional): API Key to log in with; used in conjunction with ``secretkey``. Defaults to None.
		secretkey (str, optional): API Secret to log in with; used in conjunction with ``apikey``. Defaults to None.
		verify (bool, optional): Certificate path for requests. Defaults to False.
	
	Attributes:
		baseurl (str): Base URL for appscan
		username (str): Username to log in with; used in conjunction with ``password``. Defaults to None.
		password (str): Password to log in with; used in conjunction with ``username``. Defaults to None.
		key (str): Feature Key to log in as. Defaults to "AppscanEnterpriseUser".
		apikey (str): API Key to log in with; used in conjunction with ``secretkey``. Defaults to None.
		secretkey (str): API Secret to log in with; used in conjunction with ``apikey``. Defaults to None.
		verify (bool): Certificate path for requests. Defaults to False.
		logger (:class:`logging.Logger`): logger for pyappscan; this can be overwritten by running the following
			on an instantiated object:

				>>> app.logger.handlers = []
	"""
	def __init__(self, baseurl: str, username: str = None, password: str = None, key: str ='AppscanEnterpriseUser', apikey: str = None, secretkey: str = None, verify: bool = True):
		super(PyAppscan, self).__init__(baseurl=baseurl, username=username, password=password, key=key, apikey=apikey, secretkey=secretkey, verify=verify)

	def adduser(self, realname: str, username: str, email: str, usertype: Union[str, int]) -> Dict[str, str]:
		"""Add a user to appscan with a given usertype

		Args:
			realname (str): Real name of the user
			username (str): Desired user name of the user
			email (str): Email of the user
			usertype (Union[str, int]): Usertype ID; if string, usertype ID will be grabbed

		Raises:
			UserTypeError: Not a valid user type; valid user types can be found using
				the :meth:`usertypequery` method

		Returns:
			Dict[str, str]: Result of user submission

		References:
			:meth:`pyappscan.query.Query.usertypequery`
		"""
		try:
			if bool(int(usertype)) != True:
				usertype = self.usertypequery(usertype,'name')
				if usertype == None:
					raise UserTypeError()
				else:
					usertype=int(usertype['id'])
		except ValueError:
			usertype = int(usertype)
		r = self.login()
		dic = {
			'fullName':realname,
			'userName':username,
			'email':email,
			'userTypeId':usertype
		}
		result = r.post(self.baseurl+'consoleusers', json=dic, verify=self.verify)
		return self.returnobject(result)

	def createuser(self, realname: str, username: str, email: str, usertype: Union[str, int]) -> Dict[str, str]:
		"""Alias for :meth:`adduser` for convenience

		References:
			:meth:`adduser`
		"""
		return self.adduser(realname, username, email, usertype)

	def jobcreate(self, jobname: str, policyname: Union[str, int], foldername: Union[str, int], 
		templatename: Union[str, int], jobdesc: str = None, jobcontact: str = None, appname: Union[str, int] = None) -> Dict[str, str]:
		"""Create a folder job

		Note:
			In order to initialize a job you have to run :meth:`updatescan` to 
			specify a starting URL.

		Args:
			jobname (str): Job name
			policyname (Union[str, int]): Policy ID or Name
			foldername (Union[str, int]): Folder ID or Name
			templatename (Union[str, int]): Template ID or Name
			jobdesc (str, optional): Job Description. Defaults to None.
			jobcontact (str, optional): Job contact. Defaults to None.
			appname (Union[str, int], optional): Application ID or Name to associate 
			with the job. Defaults to None.

		Returns:
			Dict[str, str]: Results of job creation including assigned Job ID

		References:
			:meth:`updatescan`
		"""
		r = self.login()
		dic = {'name':jobname}
		try:
			if bool(int(policyname)) != True:
				dic['testPolicyId'] = self.testpolicyquery(policyname,'name')['id']
			else:
				dic['testPolicyId'] = int(policyname)
		except ValueError:
			dic['testPolicyId'] = self.testpolicyquery(policyname,'name')['id']
		try:
			if bool(int(foldername)) != True:
				dic['folderId'] = self.folderquery(foldername,'folderName')['folderId']
			else:
				dic['folderId'] = int(foldername)
		except ValueError:
			dic['folderId'] = self.folderquery(foldername,'folderName')['folderId']
		if appname != None:
			try:
				if bool(int(appname)) != True:	
					dic['applicationId'] = self.applicationquery(appname,'name')['id']
				else:
					dic['applicationId'] = int(appname)
			except ValueError:
				dic['applicationId'] = self.applicationquery(appname,'name')['id']
		if jobdesc != None:
			dic['description'] = jobdesc
		if jobcontact == None:
			dic['contact'] = self.username
		else:
			dic['contact'] = jobcontact
		try:
			if bool(int(templatename)) != True:
				templateid = str(self.templatequery(templatename,'name')['id'])
			else:
				templateid = str(templatename)
		except ValueError:
			templateid = str(self.templatequery(templatename,'name')['id'])
		result = r.post(self.baseurl+'jobs/'+templateid+'/dastconfig/createjob', json=dic, verify=self.verify)
		return self.returnobject(result)
	
	def createjob(self, jobname: str, policyname: Union[str, int], foldername: Union[str, int], 
		templatename: Union[str, int], jobdesc: str = None, jobcontact: str = None, appname: Union[str, int] = None) -> Dict[str, str]:
		"""Alias for :meth:`jobcreate`

		References:
			:meth:`jobcreate`
		"""
		return self.jobcreate(jobname, policyname, foldername, templatename, jobdesc, jobcontact, appname)

	def updatescan(self, jobid: int, data: str, datacode: str = 'url', encrypt: bool = False) -> Dict[str, Any]:
		"""Updates variables in a given job

		Args:
			jobid (int): Job ID
			data (str): Data value to send to datacode
			datacode (str, optional): Data code to apply value to on given job. Defaults to 'url'. Acceptable entries:
				* ``url``
				* ``username``
				* ``password``
				* ``method``
				* ``header``
				* ``lockout``
				* ``additionaldomains``
				* ``exclusions``
			encrypt (bool, optional): Whether or not to encrypt data; should be 
			done when datacode is ``username`` or ``password`` or when other 
			sensitive info is being uploaded, like ``header`` auth tokens. Defaults to False.

		Raises:
			DataCodeError: When ``datacode`` is not valid.

		Returns:
			Dict[str, Any]: Result of scan update
		"""
		r = self.login()
		dic={
			'scantNodeNewValue':data,
			'encryptNodeValue':encrypt
		}
		if datacode == 'url' or datacode == 'URL':
  			dic['scantNodeXpath'] = 'StartingUrl'
		elif datacode == 'username' or datacode == 'Username':
			dic['scantNodeXpath'] = 'LoginUsername'
		elif datacode == 'password' or datacode == 'Password':
			dic['scantNodeXpath'] = 'LoginPassword'
		elif datacode == 'method' or datacode == 'Method':
			dic['scantNodeXpath'] = 'LoginMethod'
		elif datacode == 'header' or datacode == 'Header' or datacode == 'headers' or datacode == 'Headers':
			dic['scantNodeXpath'] = 'CustomHeaders'
		elif datacode == 'lockout' or datacode == 'Lockout':
			dic['scantNodeXpath'] = 'AccountLockout'
		elif datacode == 'additionaldomains' or datacode == 'AdditionalDomains':
			dic['scantNodeXpath'] = 'AdditionalDomains'
		elif datacode == 'exclusions' or datacode == 'Exclusions':
			dic['scantNodeXpath'] = 'Exclusions'
		else:
			raise DataCodeError()
		result = r.post(self.baseurl+'jobs/'+str(jobid)+'/dastconfig/updatescant', json=dic, verify=self.verify)
		return self.returnobject(result)

	def schedulescan(self, jobid: int, scheduleDelta: timedelta, schedule_start: datetime = None, schedule_end: datetime = None,
		exclusionDelta: timedelta = None, exclusion_start: datetime = None, exclusion_end: datetime = None) -> Dict[str, str]:
		"""Schedule scan for recurring scanning

		Note:
			This feature is still in beta.

		Args:
			jobid (int): Job ID
			scheduleDelta (:class:`datetime.timedelta`): interval to schedule at
			schedule_start (:class:`datetime.datetime`, optional): time to start scan schedule. Defaults to None.
			schedule_end (:class:`datetime.datetime`, optional): time to end scan schedule. Defaults to None.
			exclusionDelta (:class:`datetime.timedelta`, optional): interval to exlude during. Defaults to None.
			exclusion_start (:class:`datetime.datetime`, optional): time to start exclusion schedule. Defaults to None.
			exclusion_end (:class:`datetime.datetime`, optional): time to end exclusion schedule. Defaults to None.

		Returns:
			Dict[str, str]: Schedule result
		"""
		r = self.login()

		strftime_format = "%Y/%m/%d %H:%M %p"
		schedule_string, schedule_cat = self.deltaToString(scheduleDelta)
		dic = {
			"enableSchedule": True,
			"scheduleStartDate": datetime.now().strftime(strftime_format),
			"scheduleFrequency": {},
			"enableExclusion": False,
			"exclusionStartDate": datetime.now().strftime(strftime_format),
			"exclusionFrequency": {}
		}

		dic['scheduleFrequency'][schedule_cat] = schedule_string
		if schedule_start != None:
			dic['scheduleStartDate'] = schedule_start.strftime(strftime_format)
		if schedule_end != None:
			dic['scheduleEndDate'] = schedule_end.strftime(strftime_format)

		if exclusionDelta != None:
			exclusion_string, exclusion_cat = self.deltaToString(exclusionDelta)
			dic['exclusionFrequency'] = {exclusion_cat: exclusion_string}
			dic['enableExclusion'] = True

		if exclusion_start != None:
			dic['exclusionStartDate'] = exclusion_start.strftime(strftime_format)
		if exclusion_end != None:
			dic['exclusionEndDate'] = exclusion_end.strftime(strftime_format)
		
		result = r.post(self.baseurl + 'jobs/schedulescan/' + str(jobid), json=dic, verify=self.verify)
		return self.returnobject(result)

	def updatejob(self, jobid: int, jobname: str, policyname: Union[int, str], foldername: Union[int, str], 
		templatename: Union[int, str], jobdesc: str = None, jobcontact: str = None, appname: Union[int, str]=None) -> Dict[str, str]:
		"""Update job information

		Args:
			jobid (int): Job ID
			jobname (str): Job name
			policyname (int OR str): Policy ID or Name
			foldername (int OR str): Folder ID or Name
			templatename (int OR str): Template ID or Name
			jobdesc (str, optional): Job Description. Defaults to None.
			jobcontact (str, optional): Job contact. Defaults to None.
			appname (int OR str, optional): Application ID or Name to associate 
			with the job. Defaults to None.

		Returns:
			dict: Results of job update
		"""
		r = self.login()
		dic = {'name':jobname}
		try:
			if bool(int(policyname)) != True:
				dic['testPolicyId'] = self.testpolicyquery(policyname,'name')['id']
			else:
				dic['testPolicyId'] = int(policyname)
		except ValueError:
			dic['testPolicyId'] = int(policyname)
		try:
			if bool(int(foldername)) != True:
				dic['folderId'] = self.folderquery(foldername,'folderName')['folderId']
			else:
				dic['folderId'] = int(foldername)
		except ValueError:
			dic['folderId'] = int(foldername)
		if appname != None:
			try:
				if bool(int(appname)) != True:	
					dic['applicationId'] = self.applicationquery(appname,'name')['id']
				else:
					dic['applicationId'] = int(appname)
			except ValueError:
				dic['applicationId'] = int(appname)
		if jobdesc != None:
			dic['description'] = jobdesc
		if jobcontact == None:
			dic['contact'] = self.username
		else:
			dic['contact'] = jobcontact
		Etag = r.get(self.baseurl+'jobs/'+str(jobid), verify=self.verify)
		header={'If-Match':Etag.headers['ETag']}
		result = r.put(self.baseurl+'jobs/'+str(jobid), json=dic, headers=header, verify=self.verify)
		return self.returnobject(result)
	
	def startjob(self, jobid: int, action: str = 'run') -> Dict[str, str]:
		"""Start a scan on a job

		Args:
			jobid (int): Job ID
			action (str, optional): Action to take; only exists for extensibility, as 
				the only supported action as of writing is ``run``. Defaults to 'run'.

		Returns:
			Dict[str, str]: Result of job start
		"""
		r = self.login()
		if action == 'run' or action == 'Run' or action == 'RUN':
			dic={'type':'run'}
		Etag = r.get(self.baseurl+'jobs/'+str(jobid), verify=self.verify)
		header={'If-Match':Etag.headers['ETag']}
		result = r.post(self.baseurl+'jobs/'+str(jobid)+'/actions', json=dic, headers=header, verify=self.verify)
		return self.returnobject(result)
	
	def startscan(self, jobid: int, action: str = 'run') -> Dict[str, str]:
		"""Alias for :meth:`startjob`

		References:
			:meth:`startjob`
		"""
		return self.startjob(jobid, action)
	
	def createapplication(self, appname: str, attributedict: Dict[str, str] = None, 
		appdesc: str = None, appoutdated: bool = False, copyappname: str = None) -> Dict[str, Any]:
		"""Create application from attribute dictionary

		Note that the dictionary structure is:
		dict
		----dict
		--------list
		------------dict

		Note:
			When giving a ``copyappname`` and an ``attributedict``, you will overwrite only the entries in the
			``copyappname`` app's attriute dictionary which exist in the ``attributedict``.

		Args:
			appname (str): Application name
			attributedict (dict, optional): Attribute dictionary for application; you can find this by querying :meth:`applicationattributequery`. Defaults to None.
			appdesc (str, optional): App description. Defaults to None.
			appoutdated (bool, optional): Is the app outdated. Defaults to False.
			copyappname (str, optional): Application to copy and merge attributes. Defaults to None.

		Raises:
			ValueError: Raised when neither ``attributedict`` or ``copyappname`` are given,
				although they can both be given to merge attributes

		Returns:
			Dict[str, Any]: Result of app creation, including app ID

		References:
			:meth:`pyappscan.query.Query.applicationattributequery`
		"""
		r = self.login()
		attrlist = []
		currenttime = int(time()*1000)
		if attributedict == None and copyappname == None:
			raise ValueError("You must have either an attribute dictionary or a copyappname")
		if copyappname != None:
			oldliss = self.applicationattributequery(copyappname)['attributeCollection']['attributeArray']
			attributedictcopy = {}
			if attributedict != None:
				self.attributecheck(attributedict)
				oldlisscopy = {}
				for i in oldliss:
					oldlisscopy[i['name']] = i['value']
				if type(attributedict) != list:
					for k,v in attributedict:
						oldlisscopy[k] = v
					attrlist = oldlisscopy
				else:
					for i in attributedict:
						attributedictcopy[i['name']] = i['value']
					for k,v in attributedictcopy:
						oldlisscopy[k] = v
					attrlist = oldlisscopy
				attrlist = self.dicttolistofdicts(attrlist)
			else:
				attrlist = oldliss
		else:
			self.attributecheck(attributedict)
			if type(attributedict) != list:
				attrlist = self.dicttolistofdicts(attributedict)
			else:
				attrlist = attributedict
		#dictionary struct is:
		#dict
		#    dict
		#        list
		#            dict
		secondarydic = {"attributeArray" : attrlist}
		primarydic = {
			"attributeCollection" : secondarydic,
			"name" : appname,
			"dateCreated" : currenttime,
			"lastUpdated" : currenttime
		}
		if appdesc != None:
			primarydic['description'] = appdesc
		if appoutdated == True:
			primarydic['isOutOfDate'] = 1
		elif appoutdated == False:
			primarydic['isOutOfDate'] = 0
		else:
			primarydic['isOutOfDate'] = 2
		result = r.post(self.baseurl+'applications', json=primarydic, verify=self.verify)
		return self.returnobject(result)

	def updateapplication(self, appid: int, appname: str, attributedict: Dict[str, Any] = None, 
		appdesc: str = None, appoutdated: bool = False, copyappname: str = None) -> Dict[str, Any]:
		"""Update application

		Args:
			appid (int): App ID
			appname (str): Application name
			attributedict (dict, optional): Attribute dictionary for application; you can find this by querying :meth:`applicationattributequery`. Defaults to None.
			appdesc (str, optional): App description. Defaults to None.
			appoutdated (bool, optional): Is the app outdated. Defaults to False.
			copyappname (str, optional): Application to copy and merge attributes. Defaults to None.

		Raises:
			ValueError: Raised when neither ``attributedict`` or ``copyappname`` are given,
				although they can both be given to merge attributes

		Returns:
			Dict[str, Any]: Result of app creation, including app ID
		"""
		r = self.login()
		attrlist = []
		currenttime = int(time()*1000)
		if attributedict == None and copyappname == None:
			raise ValueError("You must have either an attribute dictionary or a copyappname") 
		if copyappname != None:
			oldliss = self.applicationattributequery(copyappname)['attributeCollection']['attributeArray']
			attributedictcopy = {}
			if attributedict != None:
				self.attributecheck(attributedict)
				oldlisscopy = {}
				for i in oldliss:
					oldlisscopy[i['name']] = i['value']
				if type(attributedict) != list:
					for k,v in attributedict:
						oldlisscopy[k] = v
					attrlist = oldlisscopy
				else:
					for i in attributedict:
						attributedictcopy[i['name']] = i['value']
					for k,v in attributedictcopy:
						oldlisscopy[k] = v
					attrlist = oldlisscopy
				attrlist = self.dicttolistofdicts(attrlist)
			else:
				attrlist = oldliss
		else:
			self.attributecheck(attributedict)
			if type(attributedict) != list:
				attrlist = self.dicttolistofdicts(attributedict)
			else:
				attrlist = attributedict
		#dictionary struct is:
		#dict
		#    dict
		#        list
		#            dict
		secondarydic = {"attributeArray" : attrlist}
		primarydic = {
			"attributeCollection" : secondarydic,
			"name" : appname,
			"dateCreated" : currenttime,
			"lastUpdated" : currenttime
		}
		if appdesc != None:
			primarydic['description'] = appdesc
		if appoutdated == True:
			primarydic['isOutOfDate'] = 1
		elif appoutdated == False:
			primarydic['isOutOfDate'] = 0
		else:
			primarydic['isOutOfDate'] = 2
		if type(appid) != int:
			appid = str(self.applicationquery(appid, 'name')['id'])
		else:
			appid = str(appid)
		result = r.put(self.baseurl+'applications'+appid, json=primarydic, verify=self.verify)
		return self.returnobject(result)

	def deleteapplication(self, appid: int) -> Dict[str, str]:
		"""Delete application

		Args:
			appid (int): Application ID

		Raises:
			ValueError: raised if App ID is :class:`None`

		Returns:
			Dict[str, str]: Result of deletion
		"""
		try:
			if bool(int(appid)) != True:
				appid == self.applicationquery(appid, 'name')['id']
		except ValueError:
			if appid == None:
				raise ValueError("appid can't be NoneType")
			else:
				appid == int(appid)
		r = self.login()
		result = r.delete(self.baseurl+'applications/'+str(appid), verify=self.verify)
		return self.returnobject(result)

	def createfolder(self, name: str, parent: Union[int, str], description: str = None, contact: str = None) -> Dict[str, str]:
		"""Create a folder

		Args:
			name (str): Folder name
			parent (int OR str): Folder parent ID; if string, folder ID will be grabbed
			description (str, optional): Folder description. Defaults to None.
			contact (str, optional): Folder contact info. Defaults to None.

		Raises:
			ValueError: Parent cannot be :class:`None`

		Returns:
			dict: Result of folder creation including folder ID
		"""
		try:
			if bool(int(parent)) != True:
				parent = self.folderquery(parent, 'name')['folderId']
		except ValueError:
			if parent == None:
				raise ValueError("Parent cannot be Nonetype")
			else:
				parent == int(parent)
		r = self.login()
		dic = {
			'folderName':name,
			'parentId':parent
		}
		if description:
			dic['description'] = description
		if contact:
			dic['contact'] = contact
		else:
			dic['contact'] = self.username
		result = r.post(self.baseurl+'folders/create', json=dic, verify=self.verify)
		return self.returnobject(result)
	
	def updatefolder(self, folderid: int, name: str = None, description: str = None, contact: str = None) -> Dict[str, str]:
		"""Update folder

		Args:
			folderid (int): Folder ID
			name (str, optional): Folder name. Defaults to None.
			description (str, optional): Folder description. Defaults to None.
			contact (str, optional): Folder contact info. Defaults to None.

		Returns:
			Dict[str, str]: Result of folder update
		"""
		try:
			if bool(int(folderid)) != True:
				folderid = self.folderquery(folderid, 'name')['folderId']
		except ValueError:
			folderid = int(folderid)
		r = self.login()
		dic = {}
		if name:
			dic['folderName'] = name
		if description:
			dic['description'] = description
		if contact:
			dic['contact'] = contact
		result = r.post(self.baseurl+'folders/edit/'+str(folderid), json=dic, verify=self.verify)
		return self.returnobject(result)
	
	def deletefolder(self, folderid: int) -> Dict[str, str]:
		"""Delete folder

		Args:
			folderid (int): Folder ID

		Returns:
			dict: Result of folder deletion
		"""
		try:
			if bool(int(folderid)) != True:
				folderid = self.folderquery(folderid, 'name')['folderId']
		except ValueError:
			folderid == int(folderid)
		r = self.login()
		result = r.delete(self.baseurl+'folders/delete/'+str(folderid), verify=self.verify)
		return self.returnobject(result)
	
	def createapikey(self) -> Dict[str, str]:
		"""Create API key

		Creates an API key and Secret Key for current user to utilize instead 
		of username/password combo, also no ``key`` is needed, and can be safely
		stored in environment variables or encrypted files (still do not commit 
		these to Github)!

		Note:
			Can be run as many times as you want, but invalidates old API/Secret key combo

		Returns:
			Dict[str, str]: Dictionary containing api key and secret key
		"""
		r = self.login()
		result = r.post(self.baseurl+'account/apikey', verify=self.verify)
		return self.returnobject(result)
	
	def deleteapikey(self) -> Dict[str, str]:
		"""Deletes API key for current user

		Returns:
			Dict[str, str]: Result of api key deletion
		"""
		r = self.login()
		result = r.post(self.baseurl+'account/deleteapikey', verify=self.verify)
		return self.returnobject(result)

	def createfolderitem(self, templateid, appid, size=None, empty=False, keySet=None, values=None):
		"""Create a folder item

		Args:
			templateid (int): Template ID
			appid (App ID): Application ID
			size (int, optional): size of folder item. Defaults to None.
			empty (bool, optional): Is item empty. Defaults to False.
			keySet (str, optional): Keyset for folder item. Defaults to None.
			values (dict, optional): Value dictionary for folder item. Defaults to None.

		Raises:
			NotImplementedError: Not implemented yet
		"""
		raise NotImplementedError

	def deletefolderitem(self, folderitemid: int) -> Dict[str, str]:
		"""Delete folder item

		Args:
			folderitemid (int): Folder item ID

		Returns:
			dict: Result of folder item deletion
		"""
		r = self.login()
		result = r.delete(self.baseurl + 'folderitems/{}'.format(folderitemid), verify=self.verify)
		return self.returnobject(result)